package ru.dbotthepony.kstarbound.defs

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.ints.Int2ObjectMap
import ru.dbotthepony.kommons.io.readVector2i
import ru.dbotthepony.kommons.io.writeStruct2i
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.collect.WeightedList
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.util.random.AbstractPerlinNoise
import ru.dbotthepony.kstarbound.world.UniversePos
import java.io.DataInputStream
import java.io.DataOutputStream

@JsonFactory
data class CelestialNames(
	val systemNames: WeightedList<String> = WeightedList(),
	val systemPrefixNames: WeightedList<String> = WeightedList(),
	val systemSuffixNames: WeightedList<String> = WeightedList(),
	val planetarySuffixes: ImmutableList<String> = ImmutableList.of(),
	val satelliteSuffixes: ImmutableList<String> = ImmutableList.of(),
)

@JsonFactory
data class CelestialBaseInformation(
	val planetOrbitalLevels: Int = 1,
	val satelliteOrbitalLevels: Int = 1,
	val chunkSize: Int = 1,
	val xyCoordRange: Vector2i = Vector2i.ZERO,
	val zCoordRange: Vector2i = Vector2i.ZERO,
) {
	constructor(stream: DataInputStream, isLegacy: Boolean) : this(
		stream.readInt(),
		stream.readInt(),
		stream.readInt(),
		stream.readVector2i(),
		stream.readVector2i(),
	)

	fun write(stream: DataOutputStream, isLegacy: Boolean) {
		stream.writeInt(planetOrbitalLevels)
		stream.writeInt(satelliteOrbitalLevels)
		stream.writeInt(chunkSize)
		stream.writeStruct2i(xyCoordRange)
		stream.writeStruct2i(zCoordRange)
	}
}

@JsonFactory
data class CelestialOrbitRegion(
	val regionName: String,
	val orbitRange: Vector2i,
	val bodyProbability: Double,
	val planetaryTypes: WeightedList<String> = WeightedList(),
	val satelliteTypes: WeightedList<String> = WeightedList(),
)

@JsonFactory
data class CelestialPlanet(val parameters: CelestialParameters, val satellites: Int2ObjectMap<CelestialParameters>)

@JsonFactory
data class CelestialGenerationInformation(
	val systemProbability: Double,
	val constellationProbability: Double,
	val constellationLineCountRange: Vector2i,
	val constellationMaxTries: Int,
	val maximumConstellationLineLength: Double,
	val minimumConstellationLineLength: Double,
	val minimumConstellationMagnitude: Double,
	val minimumConstellationLineCloseness: Double,

	val systemTypes: ImmutableMap<String, System>,
	val systemTypePerlin: AbstractPerlinNoise,
	val systemTypeBins: ImmutableList<Pair<Double, String>>,

	val planetaryTypes: ImmutableMap<String, Planet>,
	val satelliteTypes: ImmutableMap<String, Satellite>,
) {
	init {
		systemTypeBins.forEach {
			require(it.second.isBlank() || it.second in systemTypes) { "System type is present in systemTypeBins, but not in systemTypes: ${it.second}" }
		}
	}

	@JsonFactory
	data class System(
		var typeName: String = "",
		/**
		 * This was present in original code, but never implemented
		 */
		val constellationCapable: Boolean = true,
		val baseParameters: JsonObject = JsonObject(),
		val variationParameters: ImmutableList<JsonObject> = ImmutableList.of(),
		val orbitRegions: ImmutableList<CelestialOrbitRegion> = ImmutableList.of(),
	)

	@JsonFactory
	data class Planet(
		var typeName: String = "",
		val satelliteProbability: Double = 0.0,
		val maxSatelliteCount: Int? = null,
		val baseParameters: JsonObject = JsonObject(),
		val variationParameters: ImmutableList<JsonObject> = ImmutableList.of(),
		val orbitParameters: JsonObject = JsonObject(),
	)

	@JsonFactory
	data class Satellite(
		var typeName: String = "",
		val baseParameters: JsonObject = JsonObject(),
		val variationParameters: ImmutableList<JsonObject> = ImmutableList.of(),
		val orbitParameters: ImmutableMap<String, ImmutableList<JsonObject>> = ImmutableMap.of(),
	)

	init {
		planetaryTypes.entries.forEach {
			it.value.typeName = it.key
		}

		systemTypes.entries.forEach {
			it.value.typeName = it.key
		}

		satelliteTypes.entries.forEach {
			it.value.typeName = it.key
		}
	}
}

@JsonFactory
data class CelestialParameters(val coordinate: UniversePos, val seed: Long, val name: String, val parameters: JsonObject)
